/*
 *  Copyright (C) 2008 Patrik Gornicz, Gornicz_P (at) hotmail (dot) com.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <bear/debug.h>
#include <bear/Timer.h>
using namespace bear;

#include <iostream>
using std::cout;
using std::endl;

#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>

#include "handleSignal.h"

#include "game.h"

#include "graphics/graphics.h"
#include "input/inputManager.h"
#include "config/config.h"

/// ***** Private Method Headers *****

static void mainInit();
static void mainClean();

static void updatesInit();
static void updatesClean();

static void drawInit();
static void drawClean();

static void run();

static void updateFPSCounters();

static void handleInput();
static void update(float);
static void draw();

static int startUpdateThread(void*);
static int startDrawThread(void*);

/// ***** Private Variables *****

// variable used to determine if it is time to shutdown
static bool s_bIsRunning;

// Minimum possible wait times by used library
static const int s_iMinWaitMilli = 20;
static const int s_iMinWaitMicro = s_iMinWaitMilli * 1000;

/* Values for the main game loop
 *
 * s_iTargetUPS     := the amount of updates that is wanted in one second
 * s_fAccUpdateWait := the accumulated wait time for the update sleeps
 *
 * s_iUPS           := updates per second for the last second
 * s_iFPS           := frames per second for the last second
 * s_iUpdateCount   := counts this seconds updates
 * s_iDrawCount     := counts this seconds draws
 * s_micLastSecond  := stores the time of the last second, used for ups and fps
 */
static int      s_iTargetUPS     = 100;
static float    s_fAccUpdateWait = 0;

static int      s_iTargetFPS     = 100;
static float    s_fAccDrawWait   = 0;

static int s_iUPS, s_iFPS;
static int s_iUpdateCount, s_iDrawCount;

static Ticks s_micLastSecond;
static Timer s_timer;

static const float s_fGameStep    = 10;

static bool s_bUpdateInitialized  = false;
static bool s_bDrawInitialized    = false;

/// ***** MAIN Method *****
int main(int /*argc*/, char** /*args*/)
{
    mainInit();
    run();
    mainClean();

    return 0;
}

/// ***** Initializers/Cleaners *****

void mainInit()
{
    installSignal();

    debug::init();

    s_timer.init();
}
void mainClean()
{
    s_timer.fini();

    debug::fini();
}

void updatesInit()
{
    while(!s_bDrawInitialized)
        SDL_Delay(100);

    game::init();
    DPF(0, "Game initialized");

    input::init();
    DPF(0, "Input initialized");

    cfg::init();
    DPF(0, "Configs initialized");

    DPF(0, "Initialization Complete");

    s_bUpdateInitialized = true;
}
void updatesClean()
{
    DPF(0, "Update Thread Cleaning");

    cfg::clean();

    input::clean();

    game::clean();
}

void drawInit()
{
    graphics::init();
    DPF(0, "Graphics initialized");

    s_bDrawInitialized = true;

    while(!s_bUpdateInitialized)
        SDL_Delay(100);
}
void drawClean()
{
    DPF(0, "Draw Thread Cleaning");

    graphics::clean();
}

/// ***** Private Methods *****

void run()
{
    s_bIsRunning = true;

    SDL_Thread* s_pUpdatesThread   = SDL_CreateThread(startUpdateThread,   NULL);
    SDL_Thread* s_pDrawThread      = SDL_CreateThread(startDrawThread,     NULL);

    SDL_WaitThread(s_pUpdatesThread,   NULL);
    SDL_WaitThread(s_pDrawThread,      NULL);
}

void updateFPSCounters()
{
    // Check if a second has passed to recalculate UPS and FPS
    if (s_timer.query() - s_micLastSecond >= 1000000)
    {
        s_iUPS = s_iUpdateCount;
        s_iFPS = s_iDrawCount;

        // NOT thread safe, but they're estimates anyways
        s_iUpdateCount -= s_iUPS;
        s_iDrawCount -= s_iFPS;

        s_micLastSecond = s_timer.query();

        if(cfg::showFPS())
        {
            cout << "fps:\t" << s_iFPS << endl;
        }
        if(cfg::showUPS())
        {
            cout << "ups:\t" << s_iUPS << endl;
        }
    }
}

void handleInput()
{
    input::update();

    cfg::handleInput();
    game::handleInput();

    if(cfg::endGame())
        s_bIsRunning = false;
}

void update(float fTimeStep)
{
    game::update(fTimeStep);
    s_iUpdateCount++;
}

void draw()
{
    game::draw();

    SDL_PumpEvents(); // has to be on the Draw thread for the Windows API

    SDL_GL_SwapBuffers();

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    s_iDrawCount++;
}

int startUpdateThread(void*)
{
    updatesInit();

    s_micLastSecond = s_timer.query();

    while(s_bIsRunning)
    {
        Timer timer;
        timer.init();
            handleInput();
            update(s_fGameStep);

            updateFPSCounters();
        const Ticks ticks = timer.query();
        timer.fini();

        float wait = (1000000.0 / s_iTargetUPS - ticks.microseconds());
        s_fAccUpdateWait += 0 < wait ? wait : 0;

        if(s_iMinWaitMicro < s_fAccUpdateWait)
        {
            int iWaits = (int)(s_fAccUpdateWait / s_iMinWaitMicro);
            s_fAccUpdateWait -= iWaits * s_iMinWaitMicro;
            SDL_Delay(iWaits * s_iMinWaitMilli);
        }
    }

    updatesClean();

    return 0;
}

int startDrawThread(void*)
{
    drawInit();

    while(s_bIsRunning)
    {
        Timer timer;
        timer.init();
            draw();
        const Ticks ticks = timer.query();
        timer.fini();

        float wait = (1000000.0 / s_iTargetFPS - ticks.microseconds());
        s_fAccDrawWait += 0 < wait ? wait : 0;

        if(s_iMinWaitMicro < s_fAccDrawWait)
        {
            int iWaits = (int)(s_fAccDrawWait / s_iMinWaitMicro);
            s_fAccDrawWait -= iWaits * s_iMinWaitMicro;
            SDL_Delay(iWaits * s_iMinWaitMilli);
        }
    }

    drawClean();

    return 0;
}
